/*
===============================================================================
 Name        : MP3_16a.c
 Author      : $(author)
 Version     :
 Copyright   : $(copyright)
 Description : main definition
===============================================================================
*/

#ifdef __USE_CMSIS
#include "LPC17xx.h"
#endif

#include <cr_section_macros.h>
#include <string.h>
#include "i2s.h"
// mp3codec
#include "mp3dec.h"
#include "coder.h"
// FatFS
#include "ff.h"
#include "ffconf.h"
#include "usbhost_inc.h"
#include "Storage.h"
// 自己／デバッグ
#include "uart.h"

#define SAMPLE_FREQ 			44100	// サンプリング周波数の初期値


/*----------------------------------------------------------------------------
  RAM allocation
 *----------------------------------------------------------------------------*/
/* SRAM allocation:
 *
 * IRAM: 0x10000000 - 0x10007FFF [32 kBytes] for data, bss & stack.
 * AHB0: 0x2007C000 - 0x2007DFFF [ 8 kBytes] for READBUF
 * AHB0: 0x2007E000 - 0x2007EFFF [ 4 kBytes] for USBHost
 * AHB1: 0x20080000 - 0x20083FFF [16 kBytes] for AUDIOBUF
*/

/* USBHostLite has a low file read speed (around 9ms/8kB), it might be
 * necessary to reduce USB buffer size to 4kB to ensure smooth playback.
 * However, for the SD card the file read speed is fast enough, so the
 * buffer can be 8kB for sure.
 */
#define  READBUF_BASEADDR	0x2007C000
#define  READBUF_SIZE_USB  (1024*8)
#define  READBUF_SIZE_SD  (1024*8)

#define  AUDIOBUF_BASEADDR	0x20080000
#define  AUDIOBUF_SIZE		(0x1200)	/* 4608 byte (4.5KB) */

/* Allocate 2 or 3 audio buffer to store PCM after decoding */
#define  MAX_AUDIOBUF_NUM	3
#if (MAX_AUDIOBUF_NUM < 2) || (MAX_AUDIOBUF_NUM > 3)
    #error Invalid MAX_AUDIOBUF_NUM setting.
#endif

/*----------------------------------------------------------------------------
  MP3
 *----------------------------------------------------------------------------*/
enum {AUDIOBUF0=0, AUDIOBUF1=1, AUDIOBUF2};
typedef struct
{
	uint32_t BaseAddr;  /* base address for the audio buf */
	int32_t Size;		/* size of the audio buf */
	uint8_t Empty;		/* buf status, 1-> empty */
} AUDIOBUF;


/*----------------------------------------------------------------------------
  FAT
 *----------------------------------------------------------------------------*/
FILINFO Finfo;
FATFS Fatfs[_VOLUMES];		/* File system object for each logical drive */
FIL File1;			        /* File objects */
uint8_t Drives;				/* Drives mask */

/*----------------------------------------------------------------------------
  I2S and DMA
 *----------------------------------------------------------------------------*/

//volatile uint8_t I2S_DMA_Done = 0;
volatile uint8_t DMA_Buf_Empty = 0;

typedef unsigned short		dacout_s;
dacout_s *dac_s, *dac_e;

/*----------------------------------------------------------------------------
  MP3 variables
 *----------------------------------------------------------------------------*/

/* Play from audio buf 0, then move to the next buf. */
volatile uint8_t PlayingBufIdx = AUDIOBUF0;
volatile uint8_t DecodeBufIdx = AUDIOBUF0;

HMP3Decoder hMP3Decoder;

MP3FrameInfo mp3FrameInfo;
uint8_t *readBuf, *readPtr;
int bytesLeft, outOfData, eofReached;
uint32_t startTime, endTime, totalDecTime, nFrames;
uint32_t totalReadFileTime, nReadFileBlkCnt;

uint32_t pcmframe_size;
float audioSecs;
unsigned int samplerate;	// サンプルレートの退避

AUDIOBUF audiobuf[MAX_AUDIOBUF_NUM] = {
	{AUDIOBUF_BASEADDR, -1, 1},
	{AUDIOBUF_BASEADDR+AUDIOBUF_SIZE, -1, 1},
#if MAX_AUDIOBUF_NUM==3
	{AUDIOBUF_BASEADDR+AUDIOBUF_SIZE*2, -1, 1},
#endif
};

/*----------------------------------------------------------------------------
  others
 *----------------------------------------------------------------------------*/

#define DEBUG_MSG   DebugWriteln

static volatile uint32_t Timer;  /* Performance timer (1kHz increment):使ってない */

/*----------------------------------------------------------------------------
  local functions
 *----------------------------------------------------------------------------*/
void 	InitApp();
void 	SysTick_Off();
void 	SysTick_On();
uint8_t MP3_DecodeFrame (void);
void    MP3_AddAudioBuf (uint32_t len);
void    MP3_DisplayFormat(void);
uint8_t Mem_Init (void);
void    Mem_DeInit (void);
uint8_t MP3_Playback (void);
int 	PlayOneMp3(const char *mp3file, uint16_t dirno, uint8_t trackno);

/*----------------------------------------------------------------------------
  Functions
 *----------------------------------------------------------------------------*/
/*
 * SysTickタイマー　オフ
 */
void SysTick_Off(){
	SysTick->CTRL &= ~SysTick_CTRL_ENABLE_Msk;
}

/*
 * SysTickタイマー　オン
 */
void SysTick_On(){
	SysTick->CTRL |= SysTick_CTRL_ENABLE_Msk;
}

/* SysTick Interrupt Handler (1ms)    */
void SysTick_Handler (void)
{
	if(dac_s < dac_e){
		i2s_write(dac_s, 2);
		dac_s += 2;
	}
	if(dac_s >= dac_e){
		uint8_t next_playing_buf_idx;

        /* Data in audiobuf[PlayingBufIdx] has been sent out */
		audiobuf[PlayingBufIdx].Empty = 1;
		audiobuf[PlayingBufIdx].Size = -1;

		next_playing_buf_idx = PlayingBufIdx + 1;
	    if (next_playing_buf_idx == MAX_AUDIOBUF_NUM)
	    	next_playing_buf_idx = 0;
		if (audiobuf[next_playing_buf_idx].Empty == 0){
			PlayingBufIdx = next_playing_buf_idx;
		    dac_s = (dacout_s *)(audiobuf[PlayingBufIdx].BaseAddr);
		    dac_e = dac_s + (audiobuf[PlayingBufIdx].Size / sizeof(dacout_s));
		}
	}

}


int main(void) {
    // Force the counter to be placed into memory
    volatile static int i = 0 ;

	SystemInit();
    SystemCoreClockUpdate();
    InitApp();

    PlayOneMp3("/a.mp3", 0, 0);
    PlayOneMp3("/MSC1_in/b.mp3", 0, 0);

    // Enter an infinite loop, just incrementing a counter
    while(1) {
        i++ ;
    }
    return 0 ;
}

/*
 * 各種初期化(IO/Timer/I2S/SPI/I2C/UART/GCLD/メモリ領域)
 */
void InitApp(){
    LPC_GPIO1->FIODIR |= ((0x01 << 18) | (0x01 << 20) | (0x01 << 21) | (0x01 << 23));		// on board led
    LPC_GPIO1->FIOSET |= (0x01 << 18);
    UARTInit(1, 9600);
	DEBUG_MSG("MP3_16a start.\n\r");

    Drives = Mem_Init();
    if (Drives == 0)
    {
        DEBUG_MSG ("Memory: Init failed.\n\r");
        //LCD_PrintString(0*8, 7*15, "Mem: Init Fail", COLOR_RED, COLOR_BLACK);
        //goto loop;
        return;
    }
    LPC_GPIO1->FIOSET |= (0x01 << 20);

	i2s_init();		// i2s.c
	samplerate = SAMPLE_FREQ;
	i2s_begin(samplerate, 16);

    /* Init MP3 decoder */
	if ( (hMP3Decoder = MP3InitDecoder()) == NULL )
    {
        DEBUG_MSG("MP3 Decoder Init failed!\r\n");
        //goto loop;
        return;
    }
}


int PlayOneMp3(const char *mp3file, uint16_t dirno, uint8_t trackno){
    uint8_t res;

	res = f_open(&File1, mp3file, FA_READ | FA_OPEN_EXISTING);
	if (res != FR_OK){
		DEBUG_MSG("f_open ERR.\n\r");
		return res;
	}
	DEBUG_MSG("f_open OK.\n\r");
    /* Play */
	res = MP3_Playback();

    f_close(&File1);
	DEBUG_MSG("f_close OK.\n\r");

    return res;
}


/*----------------------------------------------------------------------------
  USB functions
 *----------------------------------------------------------------------------*/
/**
  * @brief  Initializes the USBHost .
  *
  * @param  None
  * @retval 0:OK, others: failure code
  *
  */
int USBHost_Init(void)
{
    int32_t  rc;
	uint32_t  numBlks, blkSize;
	uint8_t  inquiryResult[INQUIRY_LENGTH];

    Host_Init();               /* Initialize the lpc17xx host controller */
    rc = Host_EnumDev();       /* Enumerate the device connected  */
	if (rc != OK)
		return 1;
	rc = MS_Init( &blkSize, &numBlks, inquiryResult );
	if ((rc != OK) || (blkSize != 512))
		return 2;

	return OK;
}

/**
  * @brief  Initializes memory.
  *
  * @param  None.
  * @retval Mask of mounted drives
  *
  */
uint8_t Mem_Init (void)
{
    uint8_t retv = 0x0;

	/* Wait until there is USB flash disk inserted. */
    //LCD_PrintString(0*8, 4*15, "Connect USB MSD", COLOR_WHITE, COLOR_BLACK);
	if (USBHost_Init() == OK)
	{
		DEBUG_MSG("USBHost: Init OK.\n\r");
		//LCD_PrintString(0*8, 5*15, "USB: Init OK", COLOR_LIME, COLOR_BLACK);
		if (f_mount(/*pFatfs*/ &Fatfs[0], _T("0:"), 0) == FR_OK){
			DEBUG_MSG("USBFS: Mount OK.\n\r");
			//LCD_PrintString(0*8, 6*15, "USBFS: Mount OK", COLOR_LIME, COLOR_BLACK);
			retv |= 1;
		} else {
			DEBUG_MSG("USBFS: Mount failed.\n\r");
			//LCD_PrintString(0*8, 6*15, "USBFS: mnt fail", COLOR_RED, COLOR_BLACK);
	    }
	}
	else
	{
		DEBUG_MSG("USBHost: Init failed.\n\r");
		//LCD_PrintString(0*8, 5*15, "USB: Init fail", COLOR_RED, COLOR_BLACK);
		return retv;
	}

#ifdef SUPPORT_SD
	if (f_mount(1, &Fatfs[1]) == FR_OK) {
		DEBUG_MSG("SDFS: Mount OK.\n\r");
		LCD_PrintString(0*8, 6*15, "SDFS: Mount OK", COLOR_LIME, COLOR_BLACK);
		retv |= 2;
	} else {
		DEBUG_MSG("SDFS: Mount failed.\n\r");
		LCD_PrintString(0*8, 6*15, "SDFS: mnt fail", COLOR_RED, COLOR_BLACK);
    }
#endif

    return (retv);
}


/*----------------------------------------------------------------------------
  MP3 functions (サンプルのmp3dec_main.c から)
 *----------------------------------------------------------------------------*/

/**
  * @brief  Add an PCM frame to audio buf after decoding.
  *
  * @param  len: Specifies the frame size in bytes.
  * @retval None.
  *
  */
void MP3_AddAudioBuf (uint32_t len)
{
    /* Mark the status to not-empty which means it is available to playback. */
    audiobuf[DecodeBufIdx].Empty = 0;
    audiobuf[DecodeBufIdx].Size = len;

    /* Point to the next buffer */
    DecodeBufIdx ++;
    if (DecodeBufIdx == MAX_AUDIOBUF_NUM)
        DecodeBufIdx = 0;
}


/* Read MP3 data from file and fill in the read buffer. */
/**
  * @brief  Read data from MP3 file and fill in the Read Buffer.
  *
  * @param  readBuf: Pointer to the start of the Read Buffer
  * @param  readPtr: Pointer to the start of the left bytes in Read Buffer
  * @param  bytesLeft: Specifies the left bytes number in Read Buffer
  * @param  bytesAlign: Specifies the pad number to make sure it is aligned to 4 bytes
  * @retval Actual number of bytes read from file.
  *
  */
int MP3_FillReadBuffer(uint8_t *readBuf, uint8_t *readPtr, uint32_t bytesLeft, uint32_t bytesAlign)
{
	uint32_t nRead;

    /* Move the left bytes from the end to the front */
	memmove(readBuf+bytesAlign, readPtr, bytesLeft);

#ifdef SUPPORT_SD
	if(MP3List[CurrentMP3].drive == 1)
	{
        f_read(&File1, (void *)(readBuf+bytesLeft+bytesAlign), READBUF_SIZE_SD-bytesLeft-bytesAlign, &nRead);
		/* zero-pad to avoid finding false sync word after last frame (from old data in readBuf) */
		if (nRead < READBUF_SIZE_SD - bytesLeft - bytesAlign)
			memset(readBuf+bytesAlign+bytesLeft+nRead, 0, READBUF_SIZE_SD-bytesAlign-bytesLeft-nRead);
	}
#endif
#ifdef SUPPORT_USB
#if !defined(MP3_16)
	if(MP3List[CurrentMP3].drive == 0)
	{
#endif
		f_read(&File1, (void *)(readBuf+bytesLeft+bytesAlign), READBUF_SIZE_USB-bytesLeft-bytesAlign, &nRead);
		/* zero-pad to avoid finding false sync word after last frame (from old data in readBuf) */
		if (nRead < READBUF_SIZE_USB - bytesLeft - bytesAlign)
			memset(readBuf+bytesAlign+bytesLeft+nRead, 0, READBUF_SIZE_USB-bytesAlign-bytesLeft-nRead);
#if !defined(MP3_16)
	}
#endif
#endif

	return nRead;
}

/**
  * @brief  Decode a frame.
  *
  * @param  None.
  * @retval 0: success, 1: terminated.
  *
  */
uint8_t MP3_DecodeFrame ()
{
    uint8_t word_align, frame_decoded;
    int nRead, offset, err;

    frame_decoded = 0;
    word_align = 0;
    nRead = 0;
    offset = 0;

	do
	{
		/* somewhat arbitrary trigger to refill buffer - should always be enough for a full frame */
		if (bytesLeft < 2*MAINBUF_SIZE && !eofReached)
        {
            /* Align to 4 bytes */
    		word_align = (4-(bytesLeft&3)) & 3;

            Timer = 0;
            /* Fill read buffer */
    	    nRead = MP3_FillReadBuffer(readBuf, readPtr, bytesLeft, word_align);
    		totalReadFileTime += Timer;
    		nReadFileBlkCnt++;

    		bytesLeft += nRead;
    		readPtr = readBuf + word_align;
    		if (nRead == 0)
    		{
    			eofReached = 1;	/* end of file */
    			outOfData = 1;
    		}
		}

		/* find start of next MP3 frame - assume EOF if no sync found */
		offset = MP3FindSyncWord(readPtr, bytesLeft);
		if (offset < 0) {
			readPtr = readBuf;
			bytesLeft = 0;
			continue;
		}
		readPtr += offset;
		bytesLeft -= offset;

		//simple check for valid header
		if(((*(readPtr+1) & 24) == 8) || ((*(readPtr+1) & 6) != 2) || ((*(readPtr+2) & 240) == 240) || ((*(readPtr+2) & 12) == 12) || ((*(readPtr+3) & 3) == 2))
		{
			readPtr += 1;		//header not valid, try next one
			bytesLeft -= 1;
			continue;
		}

        Timer = 0;
		err = MP3Decode(hMP3Decoder, &readPtr, &bytesLeft, (short *)audiobuf[DecodeBufIdx].BaseAddr, 0);
		totalDecTime += Timer;
		nFrames++;
		if(err == -6)
		{
			readPtr += 1;
			bytesLeft -= 1;
			continue;
		}

		if (err) {
			/* error occurred */
			switch (err) {
			case ERR_MP3_INDATA_UNDERFLOW:
				outOfData = 1;
				break;
			case ERR_MP3_MAINDATA_UNDERFLOW:
				/* do nothing - next call to decode will provide more mainData */
				break;
			case ERR_MP3_FREE_BITRATE_SYNC:
			default:
				outOfData = 1;
				break;
			}
		} else {
			/* no error */
			MP3GetLastFrameInfo(hMP3Decoder, &mp3FrameInfo);
			frame_decoded = 1;
		}

	} while (!frame_decoded && !outOfData);

    if (outOfData == 1) return 0x1; /* Decoder terminated */
    else                return 0x0;   /* Decoder success. */

}

/**
  * @brief  Playback MP3 file.
  *
  * @param  None
  * @retval 0: success, none-zero: failure code
  *         1: failed to decode the first frame
  *         2: Unsupported MP3 format
  *
  */
uint8_t MP3_Playback (void)
{
    /* Init variables */
	bytesLeft = 0;
	outOfData = 0;
	eofReached = 0;
	readPtr = readBuf = (uint8_t *)(READBUF_BASEADDR);
	totalDecTime = 0;
	nFrames = 0;
	nReadFileBlkCnt = 0;
	totalReadFileTime = 0;
	pcmframe_size = 0;

    /* Decode and playback from audio buf 0 */
#if defined(MP3_16)
    SysTick_Off();
    dac_s = (dacout_s *)(audiobuf[PlayingBufIdx].BaseAddr);
    dac_e = dac_s;
#endif
    DecodeBufIdx = PlayingBufIdx = AUDIOBUF0;
    audiobuf[AUDIOBUF0].Empty = audiobuf[AUDIOBUF1].Empty = 1;

    /* Decode the first frame to get the frame format */
    if (MP3_DecodeFrame() == 0)
    {
        pcmframe_size = (mp3FrameInfo.bitsPerSample / 8) * mp3FrameInfo.outputSamps;
        MP3_AddAudioBuf (pcmframe_size);
    }
    else
    {
        DEBUG_MSG("\n\rFailed to decode the 1st frame, try to increase READBUF_SIZE.\n\r");
        return 0x1;  /* fail to decode the first frame */
    }

    /* Display the frame format */
    MP3_DisplayFormat();

    /* configure appropriate bit rate for each mp3, otherwise,
    the mp3 will be too fast or slow with default setting */
#if defined(MP3_16)
    if (samplerate != mp3FrameInfo.samprate){	// 設定済みsamplerateと異なる時だけi2s_beginを実行する(POPノイズ対策)
		i2s_begin(mp3FrameInfo.samprate, 16);
		samplerate = mp3FrameInfo.samprate;
    } else {
    	i2s_start();
    }
	if (SysTick_Config(SystemCoreClock / mp3FrameInfo.samprate - 1)) {
	}
#else
	if (I2S_ConfigBitrate(mp3FrameInfo.nChans, mp3FrameInfo.bitsPerSample, mp3FrameInfo.samprate))
	{
		DEBUG_MSG("Unsupported MP3 format!\n\r");
        return 0x2; /* Unsupported MP3 format */
	}
#endif

#if defined(MP3_16)
	/* Start I2S : Send the first decoded frame to I2S */
    dac_e = dac_s + (audiobuf[PlayingBufIdx].Size / 2);
    SysTick_On();
#else
    /* Start I2S+DMA to send the first decoded frame */
    I2S_Start();
    DMA_SendData((uint32_t *)audiobuf[AUDIOBUF0].BaseAddr, audiobuf[AUDIOBUF0].Size);
    I2S_DMA_Done = 0;
    DMA_Buf_Empty = 0;
#endif

    /* Then deocder all the left frames */
    do
    {
        /* wait for playback is complete to release buf */
        while (audiobuf[DecodeBufIdx].Empty == 0)
        {
            /* Decoder is faster than play, all buf is full */
#if defined(MP3_16)
#else
			decode_faster_flag ^= 0x1;
			if (decode_faster_flag)  LED_On (LED_DEC_FASTER);
			else                     LED_Off (LED_DEC_FASTER);
#endif
        }

        /* Decoder one frame */
        if (MP3_DecodeFrame() == 0)
        {
			pcmframe_size = (mp3FrameInfo.bitsPerSample / 8) * mp3FrameInfo.outputSamps;
	   		MP3_AddAudioBuf (pcmframe_size);
        }
    } while ((outOfData == 0) && (DMA_Buf_Empty ==0));

	/* Playback complete */
#if defined(MP3_16)
    SysTick_Off();
    i2s_end();
#else
    WAIFORDMADONE;
	WAIFORI2SDONE;
	I2S_Stop();
#endif

    if (outOfData == 1 || (DMA_Buf_Empty==1))
    {

    #if 1
    	MP3GetLastFrameInfo(hMP3Decoder, &mp3FrameInfo);
    	audioSecs = ((float)nFrames * mp3FrameInfo.outputSamps) / ( (float)mp3FrameInfo.samprate * mp3FrameInfo.nChans);
#if defined(MP3_16)
#else
    	DEBUG_MSG("Statistics: \n\r");
    	DEBUG_MSG("    Decode: %d frames, %d ms (%d ms/frame)\n\r", nFrames, totalDecTime, totalDecTime/nFrames);
    	DEBUG_MSG("    Audio duration(min:sec): %02d:%02d\n\r",  (uint32_t)audioSecs/60, (uint32_t)audioSecs%60);
#endif
#ifdef SUPPORT_USB
#if defined(MP3_16)
#else
    	if(MP3List[File1.fs? CurrentMP3 : PreviousMP3].drive == 0)
    		DEBUG_MSG("    File read speed: %d ms/%dkB\n\r",  totalReadFileTime/nReadFileBlkCnt, READBUF_SIZE_USB>>10);
#endif
#endif
#ifdef SUPPORT_SD
    	if(MP3List[File1.fs? CurrentMP3 : PreviousMP3].drive == 1)
     	DEBUG_MSG("    File read speed: %d ms/%dkB\n\r",  totalReadFileTime/nReadFileBlkCnt, READBUF_SIZE_SD>>10);
#endif
    #endif
    }

    return 0;
}

/**
  * @brief  Displays the format information in structure mp3FrameInfo.
  *
  * @param  None
  * @retval None
  *
  */
void MP3_DisplayFormat(void)
{
#if defined(MP3_16)
#else
	DEBUG_MSG("\n\rmp3FrameInfo:\n\r");
	DEBUG_MSG("    bitrate: %d bps\r\n", mp3FrameInfo.bitrate);
	DEBUG_MSG("    nChans: %d \r\n", mp3FrameInfo.nChans);
	DEBUG_MSG("    samprate: %d \r\n", mp3FrameInfo.samprate);
	DEBUG_MSG("    bitsPerSample: %d \r\n", mp3FrameInfo.bitsPerSample);
	DEBUG_MSG("    outputSamples: %d \r\n", mp3FrameInfo.outputSamps);
	DEBUG_MSG("    layer: %d \r\n", mp3FrameInfo.layer);
	DEBUG_MSG("    version: %d \r\n", mp3FrameInfo.version);
#endif
}

